{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "1ca56c50",
   "metadata": {},
   "source": [
    "# Outlook\n",
    "\n",
    "In this notebook, using BBRL, we code a version of the DQN algorithm with a\n",
    "replay buffer and a target network, using the AutoReset approach.\n",
    "\n",
    "To understand this code, you need to know more about \n",
    "[the BBRL interaction model](https://github.com/osigaud/bbrl/blob/master/docs/overview.md)\n",
    "Then you should run [a didactical example](https://github.com/osigaud/bbrl/blob/master/docs/notebooks/03-multi_env_autoreset.student.ipynb)\n",
    "to see how agents interact in BBRL when autoreset=True.\n",
    "\n",
    "The DQN algorithm is explained in [this\n",
    "video](https://www.youtube.com/watch?v=CXwvOMJujZk) and you can also read [the\n",
    "corresponding slides](http://pages.isir.upmc.fr/~sigaud/teach/dqn.pdf).\n",
    "\n",
    "In this notebook, we will modify the previous notebook:\n",
    "\n",
    "- to use a replay buffer and an environment that resets\n",
    "- to use a target network for $Q$\n",
    "- to use a better estimation for the maximum (Double-DQN)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3fd5978b",
   "metadata": {},
   "source": [
    "## Installation and Imports\n",
    "\n",
    "### Installation\n",
    "\n",
    "The BBRL library is [here](https://github.com/osigaud/bbrl).\n",
    "\n",
    "We use OmegaConf to that makes it possible that by just defining the `def\n",
    "run_dqn(cfg):` function and then executing a long `params = {...}` variable at\n",
    "the bottom of this colab, the code is run with the parameters without calling\n",
    "an explicit main.\n",
    "\n",
    "More precisely, the code is run by calling\n",
    "\n",
    "`config=OmegaConf.create(params)`\n",
    "\n",
    "`run_dqn(config)`\n",
    "\n",
    "at the very bottom of the colab, after starting tensorboard.\n",
    "\n",
    "Below, we import standard python packages, pytorch packages and gymnasium\n",
    "environments."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2ccb2675",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Installs the necessary Python and system libraries\n",
    "try:\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "except ModuleNotFoundError as e:\n",
    "    get_ipython().run_line_magic(\"pip\", \"install easypip\")\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "\n",
    "easyinstall(\"bbrl>=0.2.2\")\n",
    "easyinstall(\"swig\")\n",
    "easyinstall(\"bbrl_gymnasium>=0.2.0\")\n",
    "easyinstall(\"bbrl_gymnasium[box2d]\")\n",
    "easyinstall(\"bbrl_gymnasium[classic_control]\")\n",
    "easyinstall(\"tensorboard\")\n",
    "easyinstall(\"moviepy\")\n",
    "easyinstall(\"box2d-kengz\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4c7c3825",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import os\n",
    "import sys\n",
    "from pathlib import Path\n",
    "import math\n",
    "\n",
    "from moviepy.editor import ipython_display as video_display\n",
    "import time\n",
    "from tqdm.auto import tqdm\n",
    "from typing import Tuple, Optional\n",
    "from functools import partial\n",
    "\n",
    "from omegaconf import OmegaConf\n",
    "import torch\n",
    "import bbrl_gymnasium\n",
    "\n",
    "import copy\n",
    "from abc import abstractmethod, ABC\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from time import strftime\n",
    "\n",
    "OmegaConf.register_new_resolver(\n",
    "    \"current_time\", lambda: strftime(\"%Y%m%d-%H%M%S\"), replace=True\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "191dba29",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Imports all the necessary classes and functions from BBRL\n",
    "from bbrl.agents.agent import Agent\n",
    "from bbrl import get_arguments, get_class, instantiate_class\n",
    "\n",
    "# The workspace is the main class in BBRL, this is where all data is collected and stored\n",
    "from bbrl.workspace import Workspace\n",
    "\n",
    "# Agents(agent1,agent2,agent3,...) executes the different agents the one after the other\n",
    "# TemporalAgent(agent) executes an agent over multiple timesteps in the workspace,\n",
    "# or until a given condition is reached\n",
    "from bbrl.agents import Agents, TemporalAgent\n",
    "\n",
    "# ParallelGymAgent is an agent able to execute a batch of gymnasium environments\n",
    "# with auto-resetting. These agents produce multiple variables in the workspace:\n",
    "# ’env/env_obs’, ’env/reward’, ’env/timestep’, ’env/terminated’,\n",
    "# 'env/truncated', 'env/done', ’env/cumulated_reward’, ...\n",
    "#\n",
    "# When called at timestep t=0, the environments are automatically reset. At\n",
    "# timestep t>0, these agents will read the ’action’ variable in the workspace at\n",
    "# time t − 1\n",
    "from bbrl.agents.gymnasium import GymAgent, ParallelGymAgent, make_env, record_video\n",
    "\n",
    "# Replay buffers are useful to store past transitions when training\n",
    "from bbrl.utils.replay_buffer import ReplayBuffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "107b7b8c",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Utility function for launching tensorboard\n",
    "# For Colab - otherwise, it is easier and better to launch tensorboard from\n",
    "# the terminal\n",
    "def setup_tensorboard(path):\n",
    "    path = Path(path)\n",
    "    answer = \"\"\n",
    "    if is_notebook():\n",
    "        if get_ipython().__class__.__module__ == \"google.colab._shell\":\n",
    "            answer = \"y\"\n",
    "        while answer not in [\"y\", \"n\"]:\n",
    "            answer = input(\n",
    "                f\"Do you want to launch tensorboard in this notebook [y/n] \"\n",
    "            ).lower()\n",
    "\n",
    "    if answer == \"y\":\n",
    "        get_ipython().run_line_magic(\"load_ext\", \"tensorboard\")\n",
    "        get_ipython().run_line_magic(\"tensorboard\", f\"--logdir {path.absolute()}\")\n",
    "    else:\n",
    "        import sys\n",
    "        import os\n",
    "        import os.path as osp\n",
    "\n",
    "        print(\n",
    "            f\"Launch tensorboard from the shell:\\n{osp.dirname(sys.executable)}/tensorboard --logdir={path.absolute()}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "97acd32f",
   "metadata": {},
   "source": [
    "## Definition of agents\n",
    "\n",
    "### The critic agent\n",
    "\n",
    "The [DQN](https://daiwk.github.io/assets/dqn.pdf) algorithm is a critic only\n",
    "algorithm. Thus we just need a Critic agent (which will also be used to output\n",
    "actions) and an Environment agent. We reuse the `DiscreteQAgent` class that we\n",
    "have already explained in the previous notebook."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "3399eb39",
   "metadata": {},
   "source": [
    "The function below builds a multi-layer perceptron where the size of each layer is given in the `size` list.\n",
    "We also specify the activation function of neurons at each layer and optionally a different activation function for the final layer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ab3745ea",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch.nn as nn\n",
    "def build_mlp(sizes, activation, output_activation=nn.Identity()):\n",
    "    \"\"\"Helper function to build a multi-layer perceptron (function from $\\mathbb R^n$ to $\\mathbb R^p$)\n",
    "    \n",
    "    Args:\n",
    "        sizes (List[int]): the number of neurons at each layer\n",
    "        activation (nn.Module): a PyTorch activation function (after each layer but the last)\n",
    "        output_activation (nn.Module): a PyTorch activation function (last layer)\n",
    "    \"\"\"\n",
    "    layers = []\n",
    "    for j in range(len(sizes) - 1):\n",
    "        act = activation if j < len(sizes) - 2 else output_activation\n",
    "        layers += [nn.Linear(sizes[j], sizes[j + 1]), act]\n",
    "    return nn.Sequential(*layers)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1277fbea",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "class DiscreteQAgent(Agent):\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim):\n",
    "        super().__init__()\n",
    "        self.model = build_mlp(\n",
    "            [state_dim] + list(hidden_layers) + [action_dim], activation=nn.ReLU()\n",
    "        )\n",
    "\n",
    "    def forward(self, t, **kwargs):\n",
    "        obs = self.get((\"env/env_obs\", t))\n",
    "        q_values = self.model(obs)\n",
    "        self.set((f\"{self.prefix}q_values\", t), q_values)\n",
    "\n",
    "class ArgmaxActionSelector(Agent):\n",
    "    \"\"\"BBRL agent that selects the best action based on Q(s,a)\"\"\"\n",
    "    def forward(self, t: int, **kwargs):\n",
    "        q_values = self.get((\"q_values\", t))\n",
    "        action = q_values.argmax(1)\n",
    "        self.set((\"action\", t), action)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "db643104",
   "metadata": {},
   "source": [
    "### Creating an Exploration method"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f2f220fd",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "As Q-learning, DQN needs some exploration to prevent too early convergence.\n",
    "Here we will use the simple $\\epsilon$-greedy exploration method. The method\n",
    "is implemented as an agent which chooses an action based on the Q-values."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d128facc",
   "metadata": {},
   "outputs": [],
   "source": [
    "class EGreedyActionSelector(Agent):\n",
    "    def __init__(self, epsilon):\n",
    "        super().__init__()\n",
    "        self.epsilon = epsilon\n",
    "\n",
    "    def forward(self, t, **kwargs):\n",
    "        q_values = self.get((\"q_values\", t))\n",
    "        nb_actions = q_values.size()[1]\n",
    "        size = q_values.size()[0]\n",
    "        is_random = torch.rand(size).lt(self.epsilon).float()\n",
    "        random_action = torch.randint(low=0, high=nb_actions, size=(size,))\n",
    "        max_action = q_values.max(1)[1]\n",
    "        action = is_random * random_action + (1 - is_random) * max_action\n",
    "        action = action.long()\n",
    "        self.set((\"action\", t), action)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "41f89d9a",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Training and evaluation environments\n",
    "\n",
    "We build two environments: one for training and another one for evaluation.\n",
    "\n",
    "For training, it is more efficient to use an autoreset agent, as we do not\n",
    "want to waste time if the task is done in an environment sooner than in the\n",
    "others.\n",
    "\n",
    "By contrast, for evaluation, we just need to perform a fixed number of\n",
    "episodes (for statistics), thus it is more convenient to use a\n",
    "noautoreset agent with a set of environments and just run one episode in\n",
    "each environment. Thus we can use the `env/done` stop variable and take the\n",
    "average over the cumulated reward of all environments.\n",
    "\n",
    "See [this\n",
    "notebook](https://colab.research.google.com/drive/1Ui481r47fNHCQsQfKwdoNEVrEiqAEokh?usp=sharing)\n",
    "for explanations about agents and environment agents."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f9a2a896",
   "metadata": {},
   "outputs": [],
   "source": [
    "from typing import Tuple\n",
    "from bbrl.agents.gymnasium import make_env, GymAgent, ParallelGymAgent\n",
    "from functools import partial\n",
    "\n",
    "def get_env_agents(cfg, *, autoreset=True, include_last_state=True) -> Tuple[GymAgent, GymAgent]:\n",
    "    # Returns a pair of environments (train / evaluation) based on a configuration `cfg`\n",
    "    \n",
    "    # Train environment\n",
    "    train_env_agent = ParallelGymAgent(\n",
    "        partial(make_env, cfg.gym_env.env_name, autoreset=autoreset),\n",
    "        cfg.algorithm.n_envs, \n",
    "        include_last_state=include_last_state\n",
    "    ).seed(cfg.algorithm.seed)\n",
    "\n",
    "    # Test environment\n",
    "    eval_env_agent = ParallelGymAgent(\n",
    "        partial(make_env, cfg.gym_env.env_name), \n",
    "        cfg.algorithm.nb_evals,\n",
    "        include_last_state=include_last_state\n",
    "    ).seed(cfg.algorithm.seed)\n",
    "\n",
    "    return train_env_agent, eval_env_agent"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ab2021d3",
   "metadata": {},
   "source": [
    "### The Logger class\n",
    "\n",
    "The logger is in charge of collecting statistics during the training\n",
    "process.\n",
    "\n",
    "Having logging provided under the hood is one of the features allowing you\n",
    "to save time when using RL libraries like BBRL.\n",
    "\n",
    "In these notebooks, the logger is defined as `bbrl.utils.logger.TFLogger` so as\n",
    "to use a tensorboard visualisation (see the parameters part `params = { \"logger\":{ ...` below).\n",
    "\n",
    "Note that the BBRL Logger is also saving the log in a readable format such\n",
    "that you can use `Logger.read_directories(...)` to read multiple logs, create\n",
    "a dataframe, and analyze many experiments afterward in a notebook for\n",
    "instance. The code for the different kinds of loggers is available in the\n",
    "[bbrl/utils/logger.py](https://github.com/osigaud/bbrl/blob/master/src/bbrl/utils/logger.py)\n",
    "file.\n",
    "\n",
    "`instantiate_class` is an inner BBRL mechanism. The\n",
    "`instantiate_class`function is available in the\n",
    "[`bbrl/__init__.py`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/__init__.py)\n",
    "file."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "918a8052",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "from bbrl import instantiate_class\n",
    "\n",
    "class Logger():\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        self.logger = instantiate_class(cfg.logger)\n",
    "\n",
    "    def add_log(self, log_string, loss, steps):\n",
    "        self.logger.add_scalar(log_string, loss.item(), steps)\n",
    "\n",
    "    # A specific function for RL algorithms having a critic, an actor and an entropy losses\n",
    "    def log_losses(self, critic_loss, entropy_loss, actor_loss, steps):\n",
    "        self.add_log(\"critic_loss\", critic_loss, steps)\n",
    "        self.add_log(\"entropy_loss\", entropy_loss, steps)\n",
    "        self.add_log(\"actor_loss\", actor_loss, steps)\n",
    "\n",
    "    def log_reward_losses(self, rewards, nb_steps):\n",
    "        self.add_log(\"reward/mean\", rewards.mean(), nb_steps)\n",
    "        self.add_log(\"reward/max\", rewards.max(), nb_steps)\n",
    "        self.add_log(\"reward/min\", rewards.min(), nb_steps)\n",
    "        self.add_log(\"reward/median\", rewards.median(), nb_steps)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aec49fbc",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Setup the optimizers\n",
    "\n",
    "We use a single optimizer to tune the parameters of the actor (in the\n",
    "prob_agent part) and the critic (in the critic_agent part). It would be\n",
    "possible to have two optimizers which would work separately on the parameters\n",
    "of each component agent, but it would be more complicated because updating the\n",
    "actor requires the gradient of the critic."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1c4ea9ea",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "# Configure the optimizer over the q agent\n",
    "def setup_optimizers(cfg, q_agent):\n",
    "    optimizer_args = get_arguments(cfg.optimizer)\n",
    "    parameters = q_agent.parameters()\n",
    "    optimizer = get_class(cfg.optimizer)(parameters, **optimizer_args)\n",
    "    return optimizer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0f08dc17",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl import get_arguments, get_class\n",
    "from itertools import chain\n",
    "\n",
    "def setup_optimizer(cfg_optimizer, *agents):\n",
    "    \"\"\"Setup an optimizer for a list of agents\"\"\"\n",
    "    optimizer_args = get_arguments(cfg_optimizer)\n",
    "    parameters = [agent.parameters() for agent in agents]\n",
    "    optimizer = get_class(cfg_optimizer)(chain(*parameters), **optimizer_args)\n",
    "    return optimizer\n",
    "\n",
    "def copy_parameters(model_a, model_b):\n",
    "    \"\"\"Copy parameters from a model a to model_b\"\"\"\n",
    "    for model_a_p, model_b_p in zip(model_a.parameters(), model_b.parameters()):\n",
    "        model_b_p.data.copy_(model_a_p)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a7222ee9",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "# Learning environment\n",
    "\n",
    "To setup a common learning environment for RL algorithms, we use the `RLBase`\n",
    "class. This class:\n",
    "\n",
    "1. Initializes the environment (random seed, logger, evaluation environment)\n",
    "2. Defines a `evaluate` method that keeps the best agent so far\n",
    "3. Defines a `visualize_best` method that displays the behavior of the best agent\n",
    "\n",
    "Subclasses need to define `self.train_policy` and `self.eval_policy`, two\n",
    "BBRL agents that respectively choose actions when training and evaluating.\n",
    "\n",
    "The behavior of `RLBase` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `base_dir` defines the directory subpath used when outputing losses during\n",
    "training as well as other outputs (serialized agent, global statistics, etc.)\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `gym_env` defines the gymnasium environment, and in particular\n",
    "`gym_env.env_name` the name of the gymansium environment\n",
    "- `logger` defines what type of logger is used to log the different values\n",
    "associated with learning\n",
    "- `algorithm.eval_interval` defines the number of observed transitions between\n",
    "each evaluation of the agent"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a5dd3817",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from typing import Any\n",
    "import logging\n",
    "from abc import ABC\n",
    "from functools import cached_property\n",
    "\n",
    "class RLBase(ABC):\n",
    "    \"\"\"Base class for Reinforcement learning algorithms\n",
    "    \n",
    "    This class deals with common processing:\n",
    "\n",
    "    - defines the logger, the train and evaluation agents\n",
    "    - defines how to evaluate a policy\n",
    "    \"\"\"\n",
    "\n",
    "    #: The configuration\n",
    "    cfg: Any\n",
    "\n",
    "    #: The evaluation environment deals with the last action, and produces a new\n",
    "    # state of the environment\n",
    "    eval_env: Agent\n",
    "\n",
    "    #: The training policy\n",
    "    train_policy: Agent\n",
    "\n",
    "    #: The evaluation policy (if not defined, uses the training policy)\n",
    "    eval_policy: Agent\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        # Basic initialization\n",
    "        self.cfg = cfg\n",
    "        torch.manual_seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Sets the base directory and logger directory\n",
    "        base_dir = Path(self.cfg.base_dir)\n",
    "        self.base_dir = Path(\"outputs\") / Path(self.cfg.base_dir)\n",
    "        self.base_dir.mkdir(parents=True, exist_ok=True)\n",
    "\n",
    "        # Initialize the logger class\n",
    "        if not hasattr(cfg.logger, \"log_dir\"):\n",
    "            cfg.logger.log_dir = str(Path(\"outputs\") / \"tblogs\" / base_dir)\n",
    "        self.logger = Logger(cfg)\n",
    "\n",
    "        # Subclasses have to define the training and eval policies\n",
    "        self.train_policy = None\n",
    "        self.eval_policy = None\n",
    "\n",
    "        # Sets up the evaluation environment\n",
    "        self.eval_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name), \n",
    "            cfg.algorithm.nb_evals\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Initialize values\n",
    "        self.last_eval_step = 0\n",
    "        self.nb_steps = 0\n",
    "        self.best_policy = None\n",
    "        self.best_reward = -torch.inf\n",
    "\n",
    "        # Records the rewards\n",
    "        self.eval_rewards = []\n",
    "\n",
    "    @cached_property\n",
    "    def train_agent(self):\n",
    "        \"\"\"Returns the training agent\n",
    "\n",
    "        The agent is composed of a policy agent and the training environment.      \n",
    "        This method supposes that `self.train_policy` has been setup\n",
    "        \"\"\"\n",
    "        assert self.train_policy is not None, \"The train_policy property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.train_env, self.train_policy))\n",
    "\n",
    "    @cached_property\n",
    "    def eval_agent(self):\n",
    "        \"\"\"Returns the evaluation agent \n",
    "        \n",
    "        The agent is composed of a policy agent and the evaluation environment\n",
    "        \n",
    "        Uses `self.eval_policy` (or `self.train_policy` if not defined)\n",
    "\n",
    "        \"\"\"\n",
    "        assert self.eval_policy is not None or self.train_policy is not None, \"eval_agent property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.eval_env, self.eval_policy if self.eval_policy is not None else self.train_policy))\n",
    "\n",
    "    def evaluate(self):\n",
    "        \"\"\"Evaluate the current policy `self.eval_policy`\n",
    "        \n",
    "        Evaluation is conducted every `cfg.algorithm.eval_interval` steps, and\n",
    "        we keep a copy of the best agent so far in `self.best_policy`\n",
    "        \n",
    "        Returns True if the current policy is the best so far\n",
    "        \"\"\"\n",
    "        if (self.nb_steps - self.last_eval_step) > self.cfg.algorithm.eval_interval:\n",
    "            self.last_eval_step = self.nb_steps\n",
    "            eval_workspace = Workspace() \n",
    "            self.eval_agent(\n",
    "                eval_workspace,\n",
    "                t=0,\n",
    "                stop_variable=\"env/done\"\n",
    "            )\n",
    "            rewards = eval_workspace[\"env/cumulated_reward\"][-1]\n",
    "            self.logger.log_reward_losses(rewards, self.nb_steps)\n",
    "\n",
    "            if getattr(self.cfg, \"collect_stats\", False):\n",
    "                self.eval_rewards.append(rewards)\n",
    "\n",
    "            rewards_mean = rewards.mean()\n",
    "            if rewards_mean > self.best_reward:\n",
    "                self.best_policy = copy.deepcopy(self.eval_policy)\n",
    "                self.best_reward = rewards_mean\n",
    "                return True\n",
    "\n",
    "    def save_stats(self):\n",
    "        \"\"\"Save reward statistics into `stats.npy`\"\"\"\n",
    "        if getattr(self.cfg, \"collect_stats\", False) and self.eval_rewards:\n",
    "            data = torch.stack(self.eval_rewards, axis=-1) \n",
    "            with (self.base_dir / \"stats.npy\").open(\"wt\") as fp:\n",
    "                np.savetxt(fp, data.numpy())\n",
    "\n",
    "    def visualize_best(self):\n",
    "        \"\"\"Visualize the best agent\"\"\"\n",
    "        env = make_env(self.cfg.gym_env.env_name, render_mode=\"rgb_array\")\n",
    "        path = self.base_dir / \"best_agent.mp4\"\n",
    "        print(f\"Video of best agent recorded in {path}\")\n",
    "        record_video(env, self.best_policy, path)\n",
    "        return video_display(str(path.absolute()))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "68182746",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "The `EpochBasedAlgo` defines the environment when using replay buffers. In\n",
    "particular, it defines `self.train_env` which is the environment used for\n",
    "training. This environment uses *autoreset*, i.e. when reaching a terminal\n",
    "state, a new environment is created.\n",
    "\n",
    "`EpochBasedAlgo` stores all the transitions $(s_t, a_t, r_t, s_{t+1}, ...)$\n",
    "into `self.replay_buffer`.\n",
    "\n",
    "The behavior of `EpochBasedAlgo` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `gym_env.env_name` defines the gymnasium environment\n",
    "- `algorithm.n_envs` defines the number of parallel environments\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `algorithm.buffer_size` is the maximum number of transitions stored into the\n",
    "replay buffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e4825162",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "class EpochBasedAlgo(RLBase):\n",
    "    \"\"\"RL environment when using transition buffers\"\"\"\n",
    "    \n",
    "    train_agent: TemporalAgent\n",
    "    \n",
    "    \"\"\"Base class for RL experiments with full episodes\"\"\"\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        # We use a non-autoreset workspace\n",
    "        self.train_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name, autoreset=True), \n",
    "            cfg.algorithm.n_envs\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Configure the workspace to the right dimension\n",
    "        # Note that no parameter is needed to create the workspace.\n",
    "        self.replay_buffer = ReplayBuffer(max_size=cfg.algorithm.buffer_size)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "8d812d0e",
   "metadata": {},
   "source": [
    "`iter_replay_buffers` provides an easy access to the replay buffer when\n",
    "learning. Its behavior depends on several configuration values:\n",
    "\n",
    "- `cfg.algorithm.max_epochs` defines the number of times the agent is used to\n",
    "collect transitions\n",
    "- `cfg.algorithm.learning_starts` defines the number of transitions before\n",
    "learning starts\n",
    " \n",
    "Using `iter_replay_buffers` is simple:\n",
    "\n",
    "```py\n",
    "  class MyAlgo(EpochBasedAlgo):\n",
    "      def __init__(self, cfg):\n",
    "          super().__init__(cfg)\n",
    "\n",
    "          # Define the train and evaluation policies\n",
    "          # (the agents compute the workspace `action` variable)\n",
    "          self.train_policy = ...\n",
    "          self.eval_policy = ...\n",
    "\n",
    "  rl_algo = MyAlgo(cfg)\n",
    "  for rb in iter_replay_buffers(rl_algo):\n",
    "      # rb is a workspace containing transitions\n",
    "      ...\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "48ccb39f",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def iter_replay_buffers(algo: EpochBasedAlgo):\n",
    "    \"\"\"Loop over transition buffers\"\"\"\n",
    "    train_workspace = Workspace()\n",
    "\n",
    "    epochs_pb = tqdm(range(algo.cfg.algorithm.max_epochs))\n",
    "    for epoch in epochs_pb:\n",
    "        \n",
    "        # This is the tricky part with transition buffers. The difficulty lies in the\n",
    "        # copy of the last step and the way to deal with the n_steps return.\n",
    "        #\n",
    "        # The call to `train_agent(workspace, t=1, n_steps=cfg.algorithm.n_timesteps -\n",
    "        # 1, stochastic=True)` makes the agent run a number of steps in the workspace.\n",
    "        # In practice, it calls the\n",
    "        # [`__call__(...)`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/agents/agent.py#L59)\n",
    "        # function which makes a forward pass of the agent network using the workspace\n",
    "        # data and updates the workspace accordingly.\n",
    "        #\n",
    "        # Now, if we start at the first epoch (`epoch=0`), we start from the first step\n",
    "        # (`t=0`). But when subsequently we perform the next epochs (`epoch>0`), we must\n",
    "        # not forget to cover the transition at the border between the previous epoch\n",
    "        # and the current epoch. To avoid this risk, we copy the information from the\n",
    "        # last time step of the previous epoch into the first time step of the next\n",
    "        # epoch. This is explained in more details in [a previous\n",
    "        # notebook](https://colab.research.google.com/drive/1W9Y-3fa6LsPeR6cBC1vgwBjKfgMwZvP5).\n",
    "        if epoch == 0:\n",
    "            # First run: we start from scratch\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=0, n_steps=algo.cfg.algorithm.n_steps, stochastic=True\n",
    "            )\n",
    "        else:\n",
    "            # Other runs: we copy the last step and start from there\n",
    "            train_workspace.zero_grad()\n",
    "            train_workspace.copy_n_last_steps(1)\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=1, n_steps=algo.cfg.algorithm.n_steps-1, stochastic=True\n",
    "            )\n",
    "        \n",
    "        algo.nb_steps += algo.cfg.algorithm.n_steps * algo.cfg.algorithm.n_envs\n",
    "\n",
    "        # Add transitions to buffer\n",
    "        transition_workspace = train_workspace.get_transitions()\n",
    "        algo.replay_buffer.put(transition_workspace)\n",
    "        if algo.replay_buffer.size() > algo.cfg.algorithm.learning_starts:\n",
    "            yield algo.replay_buffer\n",
    "\n",
    "        # Eval\n",
    "        epochs_pb.set_description(\n",
    "            f\"nb_steps: {algo.nb_steps}, \"\n",
    "            f\"best reward: {algo.best_reward:.2f}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "fe89d447",
   "metadata": {},
   "source": [
    "## Main training loop\n",
    "\n",
    "Note that everything about the shared workspace between all the agents is\n",
    "completely hidden under the hood. This results in a gain of productivity, at\n",
    "the expense of having to dig into the BBRL code if you want to understand the\n",
    "details, change the multiprocessing model, etc.\n",
    "\n",
    "### Agent execution\n",
    "\n",
    "This is the tricky part with BBRL, the one we need to understand in detail.\n",
    "The difficulty lies in the copy of the last step and the way to deal with the\n",
    "n_steps return.\n",
    "\n",
    "The call to `train_agent(workspace, t=1, n_steps=cfg.algorithm.n_timesteps -\n",
    "1, stochastic=True)` makes the agent run a number of steps in the workspace.\n",
    "In practice, it calls the\n",
    "[`__call__(...)`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/agents/agent.py#L59)\n",
    "function which makes a forward pass of the agent network using the workspace\n",
    "data and updates the workspace accordingly.\n",
    "\n",
    "Now, if we start at the first epoch (`epoch=0`), we start from the first step\n",
    "(`t=0`). But when subsequently we perform the next epochs (`epoch>0`), we must\n",
    "not forget to cover the transition at the border between the previous epoch\n",
    "and the current epoch. To avoid this risk, we copy the information from the\n",
    "last time step of the previous epoch into the first time step of the next\n",
    "epoch.\n",
    "\n",
    "Note that we `optimizer.zero_grad()`, `loss.backward()` and `optimizer.step()`\n",
    "lines. `optimizer.zero_grad()` is necessary to cancel all the gradients\n",
    "computed at the previous iterations"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "49d135eb",
   "metadata": {},
   "outputs": [],
   "source": [
    "class DQN(EpochBasedAlgo):\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        obs_size, act_size = self.train_env.get_obs_and_actions_sizes()\n",
    "\n",
    "        # Get the two agents (critic and target critic)\n",
    "        critic = DiscreteQAgent(obs_size, cfg.algorithm.architecture.hidden_size, act_size)\n",
    "        target_critic = copy.deepcopy(critic).with_prefix(\"target/\")\n",
    "\n",
    "        # Builds the train agent that will produce transitions\n",
    "        explorer = EGreedyActionSelector(cfg.algorithm.epsilon)\n",
    "        self.train_policy = Agents(critic, explorer)\n",
    "\n",
    "        self.eval_policy = Agents(critic, ArgmaxActionSelector())\n",
    "\n",
    "        # Creates two temporal agents just for \"replaying\" some parts\n",
    "        # of the transition buffer    \n",
    "        self.q_agent = TemporalAgent(critic)\n",
    "        self.target_q_agent = TemporalAgent(target_critic)\n",
    "\n",
    "        # Get an agent that is executed on a complete workspace\n",
    "        self.optimizer = setup_optimizers(cfg, self.q_agent)\n",
    "\n",
    "        self.last_critic_update_step = 0"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "bde8ba65",
   "metadata": {},
   "source": [
    "### Compute critic loss\n",
    "\n",
    "Detailed explanations of the function to compute the critic loss when using\n",
    "`autoreset=False` are given in [this\n",
    "notebook](http://master-dac.isir.upmc.fr/rld/rl/03-1-dqn-introduction.student.ipynb).\n",
    "The case where we use `autoreset=True` is very similar, but we need to\n",
    "specify that we use the first part of the Q-values (`q_values[0]`) for\n",
    "representing $Q(s_t,a_t)$ and the second part (`q_values[1]`) for representing\n",
    "$Q(s_{t+1},a)$, as these values are stored into a transition buffer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fd74f55a",
   "metadata": {},
   "outputs": [],
   "source": [
    "def dqn_compute_critic_loss(cfg, reward, must_bootstrap, q_values, target_q_values, action):\n",
    "    \"\"\"Compute the critic loss\n",
    "\n",
    "    :param reward: The reward $r_t$ (shape 2 x B)\n",
    "    :param must_bootstrap: The must bootstrap flag at $t+1$ (shape 2 x B)\n",
    "    :param q_values: The Q-values (shape 2 x B x A)\n",
    "    :param target_q_values: The target Q-values (shape 2 x B x A)\n",
    "    :param action: The chosen actions (shape 2 x B)\n",
    "    :return: _description_\n",
    "    \"\"\"\n",
    "\n",
    "    # Implement the DQN loss\n",
    "\n",
    "    # Adapt from the previous notebook and adapt to our case (target Q network)\n",
    "    # Don't forget that we deal with transitions (and not episodes)\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "    # Compute critic loss (no need to use must_bootstrap here since we are dealing with \"full\" transitions)\n",
    "    mse = nn.MSELoss()\n",
    "    critic_loss = mse(target, qvals)\n",
    "    return critic_loss"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "72f87693",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "## Main training loop"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d0d3b28f",
   "metadata": {},
   "outputs": [],
   "source": [
    "def run(dqn: DQN, compute_critic_loss):\n",
    "    for rb in iter_replay_buffers(dqn):\n",
    "        rb_workspace = rb.get_shuffled(dqn.cfg.algorithm.batch_size)\n",
    "\n",
    "        # The q agent needs to be executed on the rb_workspace workspace\n",
    "        dqn.q_agent(rb_workspace, t=0, n_steps=2, choose_action=False)\n",
    "        with torch.no_grad():\n",
    "            dqn.target_q_agent(rb_workspace, t=0, n_steps=2, stochastic=True)\n",
    "\n",
    "        q_values, terminated, reward, action, target_q_values = rb_workspace[\n",
    "            \"q_values\", \"env/terminated\", \"env/reward\", \"action\", \"target/q_values\"\n",
    "        ]\n",
    "\n",
    "        # Determines whether values of the critic should be propagated\n",
    "        must_bootstrap = ~terminated\n",
    "\n",
    "        # Compute critic loss\n",
    "        critic_loss = compute_critic_loss(\n",
    "            dqn.cfg, reward, must_bootstrap, q_values, target_q_values, action\n",
    "        )\n",
    "        # Store the loss for tensorboard display\n",
    "        dqn.logger.add_log(\"critic_loss\", critic_loss, dqn.nb_steps)\n",
    "\n",
    "        dqn.optimizer.zero_grad()\n",
    "        critic_loss.backward()\n",
    "        torch.nn.utils.clip_grad_norm_(dqn.q_agent.parameters(), dqn.cfg.algorithm.max_grad_norm)\n",
    "        dqn.optimizer.step()\n",
    "\n",
    "        # Update target\n",
    "        if dqn.nb_steps - dqn.last_critic_update_step > dqn.cfg.algorithm.target_critic_update:\n",
    "            dqn.last_critic_update_step = dqn.nb_steps\n",
    "            copy_parameters(dqn.q_agent, dqn.target_q_agent)\n",
    "\n",
    "        dqn.evaluate()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4925a2a9",
   "metadata": {},
   "source": [
    "## Definition of the parameters\n",
    "\n",
    "The logger is defined as `bbrl.utils.logger.TFLogger` so as to use a\n",
    "tensorboard visualisation.\n",
    "\n",
    "### Launching tensorboard to visualize the results"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b3366e3d",
   "metadata": {},
   "outputs": [],
   "source": [
    "setup_tensorboard(\"./outputs/tblogs\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "77f81978",
   "metadata": {},
   "outputs": [],
   "source": [
    "params={\n",
    "  \"base_dir\": \"${gym_env.env_name}/dqn-S${algorithm.seed}_${current_time:}\",\n",
    "  \"collect_stats\": True,\n",
    "  \"save_best\": False,\n",
    "  \"logger\":{\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 4,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"epsilon\": 0.02,\n",
    "    \"n_envs\": 8,\n",
    "    \"n_steps\": 32,\n",
    "    \"n_updates\": 32,\n",
    "    \"eval_interval\": 2000,\n",
    "    \"learning_starts\": 5000,\n",
    "    \"nb_evals\": 10,\n",
    "    \"buffer_size\": 1e6,\n",
    "    \"batch_size\": 256,\n",
    "    \"target_critic_update\": 1000,\n",
    "    \"max_epochs\": 3_000,\n",
    "    \"discount_factor\": 0.99,\n",
    "    \"architecture\":{\"hidden_size\": [64, 64]},\n",
    "  },\n",
    "  \"gym_env\":{\n",
    "    \"env_name\": \"CartPole-v1\",\n",
    "  },\n",
    "  \"optimizer\":\n",
    "  {\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  }\n",
    "}\n",
    "\n",
    "dqn = DQN(OmegaConf.create(params))\n",
    "run(dqn, dqn_compute_critic_loss)\n",
    "dqn.visualize_best()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f8809bec",
   "metadata": {},
   "source": [
    "## Coding Exercise: Double DQN (DDQN)\n",
    "\n",
    "In DQN, the same network is responsible for selecting and estimating the best\n",
    "next action (in the TD-target) and that may lead to over-estimation: the\n",
    "action which q-value is over-estimated will be chosen more often. As a result,\n",
    "training is slower.\n",
    "\n",
    "To reduce over-estimation, double q-learning (and then DDQN) was proposed. It\n",
    "decouples the action selection from the value estimation.\n",
    "\n",
    "Concretely, in DQN, the target value in the critic loss (used to update the Q\n",
    "critic) for a sample at time $t$ is defined as:\n",
    "\n",
    "$$Y^{DQN}_{t} = r_{t+1} + \\gamma{Q}\\left(s_{t+1}, \\arg\\max_{a}Q\\left(s_{t+1},\n",
    "a; \\mathbb{\\theta}_{target}\\right); \\mathbb{\\theta}_{target}\\right)$$\n",
    "\n",
    "where the target network `target_q_agent` with parameters\n",
    "$\\mathbb{\\theta}_{target}$ is used for both action selection and estimation,\n",
    "and can therefore be rewritten:\n",
    "\n",
    "$$Y^{DQN}_{t} = r_{t+1} + \\gamma \\max_{a}{Q}\\left(s_{t+1}, a;\n",
    "\\mathbb{\\theta}_{target}\\right)$$\n",
    "\n",
    "Instead, DDQN uses the online critic `q_agent` with parameters\n",
    "$\\mathbb{\\theta}_{online}$ to select the action, whereas it uses the target\n",
    "network `target_q_agent` to estimate the associated Q-values:\n",
    "\n",
    "$$Y^{DDQN}_{t} = r_{t+1} + \\gamma{Q}\\left(s_{t+1}, \\arg\\max_{a}Q\\left(s_{t+1},\n",
    "a; \\mathbb{\\theta}_{online}\\right); \\mathbb{\\theta}_{target}\\right)$$\n",
    "\n",
    "The goal in this exercise is for you to write the update method for `DDQN`."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b1cdae78",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "outputs": [],
   "source": [
    "def ddqn_compute_critic_loss(cfg, reward, must_bootstrap, q_values, target_q_values, action):\n",
    "    \"\"\"Compute the critic loss\n",
    "\n",
    "    :param reward: The reward $r_t$ (shape 2 x B)\n",
    "    :param must_bootstrap: The must bootstrap flag at $t+1$ (shape 2 x B)\n",
    "    :param q_values: The Q-values (shape 2 x B x A)\n",
    "    :param target_q_values: The target Q-values (shape 2 x B x A)\n",
    "    :param action: The chosen actions (shape 2 x B)\n",
    "    :return: the loss (a scalar)\n",
    "    \"\"\"\n",
    "    \n",
    "    # Implement the double DQN loss\n",
    "\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "\n",
    "    # Compute critic loss\n",
    "    mse = nn.MSELoss()\n",
    "    critic_loss = mse(target, qvals)\n",
    "    return critic_loss"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "90052bb9",
   "metadata": {},
   "outputs": [],
   "source": [
    "params={\n",
    "  \"base_dir\": \"${gym_env.env_name}/double-dqn-S${algorithm.seed}_${current_time:}\",\n",
    "  \"collect_stats\": True,\n",
    "  \"save_best\": False,\n",
    "  \"logger\":{\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 4,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"epsilon\": 0.02,\n",
    "    \"n_envs\": 8,\n",
    "    \"n_steps\": 32,\n",
    "    \"n_updates\": 32,\n",
    "    \"eval_interval\": 2000,\n",
    "    \"learning_starts\": 5000,\n",
    "    \"nb_evals\": 10,\n",
    "    \"buffer_size\": 1e6,\n",
    "    \"batch_size\": 256,\n",
    "    \"target_critic_update\": 1000,\n",
    "    \"max_epochs\": 3_000,\n",
    "    \"discount_factor\": 0.99,\n",
    "    \"architecture\":{\"hidden_size\": [128, 128]},\n",
    "  },\n",
    "  \"gym_env\":{\n",
    "    \"env_name\": \"CartPole-v1\",\n",
    "  },\n",
    "  \"optimizer\":\n",
    "  {\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 1e-3,\n",
    "  }\n",
    "}\n",
    "\n",
    "ddqn = DQN(OmegaConf.create(params))\n",
    "run(ddqn, ddqn_compute_critic_loss)\n",
    "ddqn.visualize_best()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "98251845",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl.stats import WelchTTest\n",
    "\n",
    "WelchTTest().plot(torch.stack(dqn.eval_rewards), torch.stack(ddqn.eval_rewards))"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_markers": "\"\"\""
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
